# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from importlib.metadata import version

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath('..'))

# -- Project information -----------------------------------------------------

project = 'Yaclog'
copyright = '2021, Andrew Cassidy'
author = 'Andrew Cassidy'
release = version('yaclog')
version = '.'.join(release.split('.')[:3])
ref = version if len(release.split('.')) == 3 else 'main'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'myst_parser',
    'sphinx_click',
    'sphinx_rtd_theme',
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx_jinja'
]

myst_heading_anchors = 2
myst_enable_extensions = [
    "colon_fence"
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

default_role = 'py:obj'

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'
html_logo = 'docs_logo.png'
html_favicon = 'favicon.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_css_files = ['css/custom.css']

# -- Options for Autodoc -----------------------------------------------------

add_module_names = False
autodoc_docstring_signature = True
autoclass_content = 'both'

autodoc_default_options = {
    'member-order': 'bysource',
    'undoc-members': True,
}



# -- Options for Intersphinx -------------------------------------------------

# This config value contains the locations and names of other projects that
# should be linked to in this documentation.

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'packaging': ('https://packaging.pypa.io/en/latest/', None),
}

jinja_globals = {
    'version': version,
    'release': release,
    'ref': ref,
}


import datetime
import os.path
import textwrap

import yaclog.changelog

log_segments = [
    '# Changelog',

    'This changelog is for testing the parser, and has many things in it that might trip it up.',

    '## [Tests]',  # 2

    '- bullet point with no section',

    '### Bullet Points',  # 4

    textwrap.dedent('''\
    - bullet point dash
    * bullet point star
    + bullet point plus
      - sub point 1
      - sub point 2
      - sub point 3'''),

    '### Blocks ##',  # 6

    '#### This is an H4',
    '##### This is an H5',
    '###### This is an H6',

    '- this is a bullet point\nit spans many lines',

    'This is\na paragraph\nit spans many lines',

    '```python\nthis is some example code\nit spans many lines\n```',

    '> this is a block quote\nit spans many lines',

    '[FullVersion] - 1969-07-20 [TAG1] [TAG2]\n-----',  # 14
    '## Long Version Name',  # 15

    '[fullVersion]: http://endless.horse\n[id]: http://www.koalastothemax.com'
]

log_text = '\n\n'.join(log_segments)

log = yaclog.Changelog()
log.preamble = '# Changelog\n\n' \
               'This changelog is for testing the parser, and has many things in it that might trip it up.'
log.links = {'id': 'http://www.koalastothemax.com'}
log.versions = [yaclog.changelog.VersionEntry(), yaclog.changelog.VersionEntry(), yaclog.changelog.VersionEntry()]

log.versions[0].name = '[Tests]'
log.versions[0].sections = {
    '': ['- bullet point with no section'],
    'Bullet Points': [
        '- bullet point dash',
        '* bullet point star',
        '+ bullet point plus\n  - sub point 1\n  - sub point 2\n  - sub point 3'],
    'Blocks': [
        '#### This is an H4',
        '##### This is an H5',
        '###### This is an H6',

        '- this is a bullet point\nit spans many lines',

        'This is\na paragraph\nit spans many lines',

        '```python\nthis is some example code\nit spans many lines\n```',

        '> this is a block quote\nit spans many lines',
    ]
}

log.versions[1].name = 'FullVersion'
log.versions[1].link = 'http://endless.horse'
log.versions[1].tags = ['TAG1', 'TAG2']
log.versions[1].date = datetime.date.fromisoformat('1969-07-20')

log.versions[2].name = 'Long Version Name'



import datetime
import os.path
import tempfile
import unittest

import yaclog
from tests.common import log, log_segments, log_text
from yaclog.changelog import VersionEntry


class TestParser(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with tempfile.TemporaryDirectory() as td:
            cls.path = os.path.join(td, 'changelog.md')
            with open(cls.path, 'w') as fd:
                fd.write(log_text)
            cls.log = yaclog.read(cls.path)

    def test_path(self):
        """Test the log's path"""
        self.assertEqual(self.path, self.log.path)

    def test_preamble(self):
        """Test the preamble at the top of the file"""
        self.assertEqual(log.preamble, self.log.preamble)

    def test_links(self):
        """Test the links at the end of the file"""
        self.assertEqual({'fullversion': 'http://endless.horse', **log.links}, self.log.links)

    def test_versions(self):
        """Test the version headers"""
        for i in range(len(self.log.versions)):
            self.assertEqual(log.versions[i].name, self.log.versions[i].name)
            self.assertEqual(log.versions[i].link, self.log.versions[i].link)
            self.assertEqual(log.versions[i].date, self.log.versions[i].date)
            self.assertEqual(log.versions[i].tags, self.log.versions[i].tags)

    def test_entries(self):
        """Test the change entries"""
        self.assertEqual(log.versions[0].sections, self.log.versions[0].sections)


class TestWriter(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with tempfile.TemporaryDirectory() as td:
            cls.path = os.path.join(td, 'changelog.md')
            log.write(cls.path)
            with open(cls.path) as fd:
                cls.log_text = fd.read()
                cls.log_segments = [line.lstrip('\n') for line in cls.log_text.split('\n\n') if line]

    def test_preamble(self):
        """Test the header information at the top of the file"""
        self.assertEqual(log_segments[0:2], self.log_segments[0:2])

    def test_links(self):
        """Test the links at the end of the file"""
        self.assertEqual(
            {'[fullversion]: http://endless.horse', '[id]: http://www.koalastothemax.com'},
            set(self.log_segments[16:18]))

    def test_versions(self):
        """Test the version headers"""
        self.assertEqual('## [Tests]', self.log_segments[2])
        self.assertEqual('## [FullVersion] - 1969-07-20 [TAG1] [TAG2]', self.log_segments[14])
        self.assertEqual('## Long Version Name', self.log_segments[15])

    def test_entries(self):
        """Test the change entries"""
        self.assertEqual(log_segments[3], self.log_segments[3])
        self.assertEqual('### Bullet Points', self.log_segments[4])
        self.assertEqual(log_segments[5], self.log_segments[5])
        self.assertEqual('### Blocks', self.log_segments[6])
        self.assertEqual(log_segments[7:14], self.log_segments[7:14])


class TestVersionEntry(unittest.TestCase):
    def test_header_name(self):
        """Test reading version names from headers"""
        headers = {
            'short': ('## Test', 'Test'),
            'with dash': ('## Test - ', 'Test'),
            'multi word': ('## Very long version name 1.0.0', 'Very long version name 1.0.0'),
            'with brackets': ('## [Test]', '[Test]'),
        }

        for c, t in headers.items():
            h = t[0]
            with self.subTest(c, h=h):
                version = VersionEntry.from_header(h)
                self.assertEqual(version.name, t[1])
                self.assertEqual(version.tags, [])
                self.assertIsNone(version.date)
                self.assertIsNone(version.link)
                self.assertIsNone(version.link_id)

    def test_header_tags(self):
        """Test reading version tags from headers"""
        headers = {
            'no dash': ('## Test [Foo] [Bar]', 'Test', ['FOO', 'BAR']),
            'with dash': ('## Test - [Foo] [Bar]', 'Test', ['FOO', 'BAR']),
            'with brackets': ('## [Test] [Foo] [Bar]', '[Test]', ['FOO', 'BAR']),
            'with brackets & dash': ('## [Test] - [Foo] [Bar]', '[Test]', ['FOO', 'BAR']),
        }

        for c, t in headers.items():
            h = t[0]
            with self.subTest(c, h=h):
                version = VersionEntry.from_header(h)
                self.assertEqual(version.name, t[1])
                self.assertEqual(version.tags, t[2])
                self.assertIsNone(version.date)
                self.assertIsNone(version.link)
                self.assertIsNone(version.link_id)

    def test_header_date(self):
        """Test reading version dates from headers"""

        headers = {
            'no dash': ('## Test 1961-04-12', 'Test',
                        datetime.date.fromisoformat('1961-04-12'), []),
            'with dash': ('## Test 1969-07-20', 'Test',
                          datetime.date.fromisoformat('1969-07-20'), []),
            'two dates': ('## 1981-07-20 1988-11-15', '1981-07-20',
                          datetime.date.fromisoformat('1988-11-15'), []),
            'single date': ('## 2020-05-30', '2020-05-30', None, []),
            'with tags': ('## 1.0.0 - 2021-04-19 [Foo] [Bar]', '1.0.0',
                          datetime.date.fromisoformat('2021-04-19'), ['FOO', 'BAR']),
        }

        for c, t in headers.items():
            h = t[0]
            with self.subTest(c, h=h):
                version = VersionEntry.from_header(h)
                self.assertEqual(version.name, t[1])
                self.assertEqual(version.date, t[2])
                self.assertEqual(version.tags, t[3])
                self.assertIsNone(version.link)
                self.assertIsNone(version.link_id)

    def test_header_noncompliant(self):
        """Test reading version that dont fit the schema, and should just be read as literals"""

        headers = {
            'no space between tags': 'Test [Foo][Bar]',
            'text at end': 'Test [Foo] [Bar] Test',
            'invalid date': 'Test - 9999-99-99',
        }

        for c, h in headers.items():
            with self.subTest(c, h=h):
                version = VersionEntry.from_header('## ' + h)
                self.assertEqual(version.name, h)
                self.assertEqual(version.tags, [])
                self.assertIsNone(version.date)
                self.assertIsNone(version.link)
                self.assertIsNone(version.link_id)


if __name__ == '__main__':
    unittest.main()



#  yaclog: yet another changelog tool
#  Copyright (c) 2022. Andrew Cassidy
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from tomlkit import dumps
from tomlkit import parse


def set_version(path, version):
    """
    Set the version string in a cargo.toml file

    :param path: path-like file location
    :param version: version string to overwrite with
    """
    with open(path, 'r+') as fp:
        toml = parse(fp.read())
        toml['package']['version'] = version
        fp.seek(0)
        fp.write(dumps(toml))
        fp.truncate()



import os.path
import unittest
import traceback

import git
from click.testing import CliRunner

import yaclog.changelog
from yaclog.cli.__main__ import cli


def check_result(runner, result, success: bool = True):
    runner.assertEqual((result.exit_code == 0), success,
                       f'\noutput: {result.output}\ntraceback: ' + ''.join(
                           traceback.format_exception(*result.exc_info)))


class TestCreation(unittest.TestCase):
    def test_init(self):
        """Test creating and overwriting a changelog"""
        runner = CliRunner()
        location = 'CHANGELOG.md'
        err_str = 'THIS FILE WILL BE OVERWRITTEN'

        with runner.isolated_filesystem():
            result = runner.invoke(cli, ['init'])
            check_result(self, result)
            self.assertTrue(os.path.exists(os.path.abspath(location)), 'yaclog init did not create a file')
            self.assertIn(location, result.output, "yaclog init did not echo the file's correct location")

            with open(location, 'r') as fp:
                self.assertEqual('# Changelog\n', fp.readline())
                self.assertEqual('\n', fp.readline())
                self.assertEqual('All notable changes to this project will be documented in this file',
                                 fp.readline().rstrip())

            with open(location, 'w') as fp:
                fp.write(err_str)

            result = runner.invoke(cli, ['init'], input='y\n')
            check_result(self, result)
            self.assertTrue(os.path.exists(os.path.abspath(location)), 'file no longer exists after overwrite')
            self.assertIn(location, result.output, "yaclog init did not echo the file's correct location")

            with open(location, 'r') as fp:
                self.assertNotEqual(fp.read(), err_str, 'file was not overwritten')

    def test_init_path(self):
        """Test creating a changelog with a non-default filename"""
        runner = CliRunner()
        location = 'A different file.md'

        with runner.isolated_filesystem():
            result = runner.invoke(cli, ['--path', location, 'init'])
            check_result(self, result)
            self.assertTrue(os.path.exists(os.path.abspath(location)), 'yaclog init did not create a file')
            self.assertIn(location, result.output, "yaclog init did not echo the file's correct location")

    def test_does_not_exist(self):
        """Test if an error is thrown when the file does not exist"""
        runner = CliRunner()

        with runner.isolated_filesystem():
            result = runner.invoke(cli, ['show'])
            check_result(self, result, False)
            self.assertIn('does not exist', result.output)


class TestTagging(unittest.TestCase):
    def test_tag_addition(self):
        """Test adding tags to versions"""
        runner = CliRunner()
        location = 'CHANGELOG.md'

        with runner.isolated_filesystem():
            in_log = yaclog.Changelog(location)
            in_log.versions = [yaclog.changelog.VersionEntry(), yaclog.changelog.VersionEntry()]

            in_log.versions[0].name = '1.0.0'
            in_log.versions[1].name = '0.9.0'
            in_log.write()

            result = runner.invoke(cli, ['tag', 'tag1'])
            check_result(self, result)

            result = runner.invoke(cli, ['tag', 'tag2', '0.9.0'])
            check_result(self, result)

            out_log = yaclog.read(location)
            self.assertEqual(out_log.versions[0].tags, ['TAG1'])
            self.assertEqual(out_log.versions[1].tags, ['TAG2'])

            result = runner.invoke(cli, ['tag', 'tag3', '0.8.0'])
            check_result(self, result, False)

    def test_tag_deletion(self):
        """Test deleting tags from versions"""
        runner = CliRunner()
        location = 'CHANGELOG.md'

        with runner.isolated_filesystem():
            in_log = yaclog.Changelog(location)
            in_log.versions = [None, None]
            in_log.versions = [yaclog.changelog.VersionEntry(), yaclog.changelog.VersionEntry()]

            in_log.versions[0].name = '1.0.0'
            in_log.versions[0].tags = ['TAG1']

            in_log.versions[1].name = '0.9.0'
            in_log.versions[1].tags = ['TAG2']
            in_log.write()

            result = runner.invoke(cli, ['tag', '-d', 'tag2', '0.8.0'])
            check_result(self, result, False)

            result = runner.invoke(cli, ['tag', '-d', 'tag3', '0.9.0'])
            check_result(self, result, False)

            result = runner.invoke(cli, ['tag', '-d', 'tag1'])
            check_result(self, result)

            out_log = yaclog.read(location)
            self.assertEqual(out_log.versions[0].tags, [])
            self.assertEqual(out_log.versions[1].tags, ['TAG2'])

            result = runner.invoke(cli, ['tag', '-d', 'tag2', '0.9.0'])
            check_result(self, result)

            out_log = yaclog.read(location)
            self.assertEqual(out_log.versions[0].tags, [])
            self.assertEqual(out_log.versions[1].tags, [])


class TestRelease(unittest.TestCase):
    def test_increment(self):
        """Test version incrementing on release"""
        runner = CliRunner()
        location = 'CHANGELOG.md'

        with runner.isolated_filesystem():
            runner.invoke(cli, ['init'])  # create the changelog
            runner.invoke(cli, ['entry', '-b', 'entry number 1'])

            result = runner.invoke(cli, ['release', '1.0.0'])
            check_result(self, result)
            self.assertEqual(yaclog.read(location).versions[0].name, '1.0.0')
            self.assertIn('1.0.0', result.output)

            runner.invoke(cli, ['entry', '-b', 'entry number 2'])

            result = runner.invoke(cli, ['release', '-p'])
            check_result(self, result)
            self.assertEqual(yaclog.read(location).versions[0].name, '1.0.1')
            self.assertIn('1.0.1', result.output)

            result = runner.invoke(cli, ['release', '-y', '-s', 2])
            check_result(self, result)
            self.assertEqual(yaclog.read(location).versions[0].name, '1.0.2')
            self.assertIn('1.0.2', result.output)

            runner.invoke(cli, ['entry', '-b', 'entry number 3'])

            result = runner.invoke(cli, ['release', '-m'])
            check_result(self, result)
            self.assertEqual(yaclog.read(location).versions[0].name, '1.1.0')
            self.assertIn('1.1.0', result.output)

            runner.invoke(cli, ['entry', '-b', 'entry number 4'])

            result = runner.invoke(cli, ['release', '-M'])
            check_result(self, result)
            self.assertEqual(yaclog.read(location).versions[0].name, '2.0.0')
            self.assertIn('2.0.0', result.output)

            runner.invoke(cli, ['entry', '-b', 'entry number 5'])

            result = runner.invoke(cli, ['release', '-Ma'])
            check_result(self, result)
            self.assertEqual(yaclog.read(location).versions[0].name, '3.0.0a1')
            self.assertIn('3.0.0a1', result.output)

            result = runner.invoke(cli, ['release', '-b'])
            check_result(self, result)
            self.assertEqual(yaclog.read(location).versions[0].name, '3.0.0b1')
            self.assertIn('3.0.0b1', result.output)

            result = runner.invoke(cli, ['release', '-r'])
            check_result(self, result)
            self.assertEqual(yaclog.read(location).versions[0].name, '3.0.0rc1')
            self.assertIn('3.0.0rc1', result.output)

            result = runner.invoke(cli, ['release', '-r'])
            check_result(self, result)
            self.assertEqual(yaclog.read(location).versions[0].name, '3.0.0rc2')
            self.assertIn('3.0.0rc1', result.output)

            result = runner.invoke(cli, ['release', '-f'])
            check_result(self, result)
            self.assertEqual(yaclog.read(location).versions[0].name, '3.0.0')
            self.assertIn('3.0.0', result.output)

            result = runner.invoke(cli, ['release', '-p', '-n'])
            check_result(self, result)
            self.assertEqual(yaclog.read(location).versions[0].name, '3.0.1')
            self.assertEqual(yaclog.read(location).versions[1].name, '3.0.0')
            self.assertIn('3.0.1', result.output)

    def test_commit(self):
        """Test committing and tagging releases"""
        runner = CliRunner()

        with runner.isolated_filesystem():
            repo = git.Repo.init(os.path.join(os.curdir, 'testing'))
            os.chdir('testing')
            repo.index.commit('initial commit')

            with repo.config_writer() as cw:
                cw.set_value('user', 'email', 'unit-tester@example.com')
                cw.set_value('user', 'name', 'unit-tester')

            runner.invoke(cli, ['init'])  # create the changelog
            runner.invoke(cli, ['entry', '-b', 'entry number 1'])

            result = runner.invoke(cli, ['release', 'Version 1.0.0', '-c'], input='y\n')
            check_result(self, result)
            self.assertIn('Created commit', result.output)
            self.assertIn('Created tag', result.output)
            self.assertIn(repo.head.commit.hexsha[0:7], result.output)
            self.assertEqual(repo.tags[0].name, '1.0.0')

    def test_cargo(self):
        """Test updating cargo.toml files"""
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open("Cargo.toml", "w") as fp:
                fp.write((
                    '[package]\n'
                    'name = "dummy"\n'
                    'version = "0.3.4"\n'
                    'authors = ["Andrew Cassidy <drewcassidy@me.com>"]\n'
                    'description = "A dummy crate used for testing yaclog"\n'
                    'keywords = ["does", "not", "exist"]\n'
                    'edition = "2018"\n'
                ))

            runner.invoke(cli, ['init'])  # create the changelog
            runner.invoke(cli, ['entry', '-b', 'entry number 1'])

            result = runner.invoke(cli, ['release', 'Version 1.0.0', '-C'])
            check_result(self, result)

            with open("Cargo.toml", "r") as fp:
                self.assertIn('version = "1.0.0"', fp.read())
                # we're just going to trust tomlkit not to mangle everything else


class TestShow(unittest.TestCase):

    # noinspection PyShadowingNames
    def setUp(self):
        self.runner = CliRunner()
        self.location = 'CHANGELOG.md'

        self.log = yaclog.Changelog()

        self.log.add_version(name='1.0.0').add_entry('- entry number 1')
        self.log.add_version(name='Version 2.0.0').add_entry('- entry number 2', 'Added')
        self.log.add_version(name='Three Point Oh').add_entry('entry number 3')
        v = self.log.add_version(name='4.0.0 "Euclid"')
        v.add_entry('- entry number 4')
        v.add_entry('- entry number 5')
        v.tags.append('TAGGED')

        self.modes = {
            'full': ([], lambda v, k: v.text(**k), '\n\n'),
            'name': (['-n'], lambda v, k: v.name, '\n'),
            'body': (['-b'], lambda v, k: v.body(**k), '\n\n'),
            'header': (['-h'], lambda v, k: v.header(**k), '\n'),
        }

    def test_show_all(self):
        """Test showing all version information"""

        with self.runner.isolated_filesystem():
            self.log.write(self.location)

            for mode, t in self.modes.items():
                with self.subTest(mode, flags=t[0]):
                    check_result(self, result := self.runner.invoke(cli, ['show', '-a'] + t[0]))
                    self.assertEqual(t[2].join([t[1](v, {'md': False}) for v in self.log.versions]),
                                     result.output.strip(), 'incorrect plaintext output')

                    check_result(self, result := self.runner.invoke(cli, ['show', '-am'] + t[0]))
                    self.assertEqual(t[2].join([t[1](v, {'md': True}) for v in self.log.versions]),
                                     result.output.strip(), 'incorrect markdown output')

    def test_show_version(self):
        with self.runner.isolated_filesystem():
            self.log.write(self.location)

            for mode, t in self.modes.items():
                with self.subTest(mode, flags=t[0]):

                    for version in self.log.versions:
                        check_result(self, result := self.runner.invoke(cli, ['show', version.name] + t[0]))
                        self.assertEqual(t[1](version, {'md': False}),
                                         result.output.strip(), 'incorrect plaintext output')

                        check_result(self, result := self.runner.invoke(cli, ['show', version.name[-5:]] + t[0]))
                        self.assertEqual(t[1](version, {'md': False}),
                                         result.output.strip(), 'incorrect plaintext output')

                        check_result(self, result := self.runner.invoke(cli, ['show', version.name, '-m'] + t[0]))
                        self.assertEqual(t[1](version, {'md': True}),
                                         result.output.strip(), 'incorrect markdown output')

                        check_result(self, result := self.runner.invoke(cli, ['show', version.name[-5:], '-m'] + t[0]))
                        self.assertEqual(t[1](version, {'md': True}),
                                         result.output.strip(), 'incorrect markdown output')


if __name__ == '__main__':
    unittest.main()



"""
Tools for parsing and manipulating markdown, including a very basic tokenizer.
"""

#  yaclog: yet another changelog tool
#  Copyright (c) 2021. Andrew Cassidy
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import re
from typing import List

bullets = '+-*'
brackets = '[]'
code_regex = re.compile(r'^```')
header_regex = re.compile(r'^(?P<hashes>#+)\s+(?P<contents>[^#]+)(?:\s+#+)?$')
li_regex = re.compile(r'^[-+*] |\d+\. ')
numbered_regex = re.compile(r'^\d+\. ')
bullet_regex = re.compile(r'^[-+*] ')
link_id_regex = re.compile(r'^\[(?P<link_id>\S*)]:\s*(?P<link>.*)')
link_def_regex = re.compile(r'\[(?P<text>.*?)]\[(?P<link_id>.*?)]')  # deferred link in the form [name][id]
link_lit_regex = re.compile(r'\[(?P<text>.*?)]\((?P<link>.*?)\)')  # literal link in the form [name](url)

setext_h1_replace_regex = re.compile(r'(?<=\n)(?P<header>[^\n]+?)\n=+[ \t]*(?=\n)')
setext_h2_replace_regex = re.compile(r'(?<=\n)(?P<header>[^\n]+?)\n-+[ \t]*(?=\n)')


def strip_link(text):
    """
    Parses and removes any links from the input string

    :param text: An input string which may be a markdown link, either literal or an ID
    :return: A tuple of (name, url, id). If the input is not a link, it is returned verbatim as the name.
    """

    if link_lit := link_lit_regex.fullmatch(text):
        # in the form [name](link)
        return link_lit['text'], link_lit['link'], None

    if link_def := link_def_regex.fullmatch(text):
        # in the form [name][id] where id is hopefully linked somewhere else in the document
        return link_def['text'], None, link_def['link_id'].lower()

    return text, None, None


def join(segments: List[str]) -> str:
    """
    Joins multiple lines of markdown by adding double newlines between them, or a single newline between list items

    :param segments: A list of strings to join
    :return: A joined markdown string
    """

    text: List[str] = []
    last_segment = ''
    for segment in segments:
        if bullet_regex.match(segment) and bullet_regex.match(last_segment):
            pass
        elif numbered_regex.match(segment) and numbered_regex.match(last_segment):
            pass
        else:
            text.append('')

        text.append(segment)

        last_segment = segment

    return '\n'.join(text).strip()


class Token:
    """A single tokenized block of markdown, consisting of one or more lines of text."""

    def __init__(self, line_no: int, lines: List[str], kind: str):
        self.line_no = line_no
        """Which line this block appears on in the original file"""

        self.lines = lines
        """The lines of text making up this block"""

        self.kind = kind
        """What kind of token this is. One of ``h[1-6]``, ``p``, ``li`` or ``code``"""

    def __str__(self):
        return f'{self.kind}: {self.lines}'


def tokenize(text: str):
    """
    Tokenize a markdown string

    The tokenizer is very basic, and only cares about the highest-level blocks
    (Headers, top-level list items, links, code blocks, paragraphs).

    :param text: input text to tokenize
    :return: A list of tokens and a dictionary of links
    """

    # convert setext-style headers
    # The extra newline is to preserve line numbers
    text = setext_h1_replace_regex.sub(r'# \g<header>\n', text)
    text = setext_h2_replace_regex.sub(r'## \g<header>\n', text)

    lines = text.split('\n')
    tokens: List[Token] = []
    links = {}

    # state variables for parsing
    block = None

    for line_no, line in enumerate(lines):
        if block == 'code':
            # this is the contents of a code block
            assert block == tokens[-1].kind, 'block state variable in invalid state!'
            tokens[-1].lines.append(line)
            if code_regex.match(line):
                block = None

        elif code_regex.match(line):
            # this is the start of a code block
            tokens.append(Token(line_no, [line], 'code'))
            block = 'code'

        elif li_regex.match(line):
            # this is a list item
            tokens.append(Token(line_no, [line], 'li'))
            block = 'li'

        elif match := header_regex.match(line):
            # this is a header
            kind = f'h{len(match["hashes"])}'
            tokens.append(Token(line_no, [line], kind))

        elif match := link_id_regex.match(line):
            # this is a link definition in the form '[id]: link'
            links[match['link_id'].lower()] = match['link']
            block = None

        elif not line or line.isspace():
            # skip empty lines and reset block
            block = None

        elif block:
            # this is a line to be added to a paragraph or list item
            assert block == tokens[-1].kind, f'block state variable in invalid state! {block} != {tokens[-1].kind}'
            tokens[-1].lines.append(line)

        else:
            # this is a new paragraph
            tokens.append(Token(line_no, [line], 'p'))
            block = 'p'

    return tokens, links
